//
// factor.cc
//
// A function to factorize a univariate polynomial with integer coefficients
// ex factorpoly( const ex &PolyIn, const symbol &x )
//
/*
 *  gTybalt Copyright (C) 2001 Roberta Marani and Stefan Weinzierl
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "factor.h"
//// used by ZZFromNumeric and NumericFromZZ:
#include <string>
#include <limits.h>
#include <stdexcept>

using namespace GiNaC;

/////////////////////////////////////////////////////////////////////////////
ex factorpoly( const ex &PolyIn, const symbol &x )
{
  ex PolyOut;
  ex IrreduxPoly;

  // check if PolyIn is a polynomial with integer coefficients
  if (!PolyIn.info(info_flags::integer_polynomial))
    { 
      //    std::cout<<"Input expression is not an integer polynomial"<<std::endl;
      return PolyIn ; 
    }

  ex Poly = PolyIn.expand();

  // extract coefficients
  NTL::ZZX NTLPoly;
  NTL::ZZ NTLcoeff;
  numeric PolyCoeff;
  ex tmpCoeff;
  for (long i=Poly.ldegree(x); i<=Poly.degree(x); ++i)
    {
      tmpCoeff = Poly.coeff(x,i);
      if(!tmpCoeff.info(info_flags::integer))
	{
	  //    std::cout<<"coefficient is not an integer"<<std::endl;
	  //    std::cout<<"PolyFactor requires a univariate integer polynomial"<<std::endl;
	  return PolyIn;
	}
      PolyCoeff=ex_to<numeric>(tmpCoeff);
      ZZFromNumeric(NTLcoeff, PolyCoeff);
      NTL::SetCoeff(NTLPoly, i, NTLcoeff);
    }

  // factorize
  NTL::vec_pair_ZZX_long factors;
  NTL::ZZ c;
  NTL::factor(c, factors, NTLPoly);
  // c is the overall coefficient
  // factors is a vector of ZZX and long pairs

  //Back to GiNaC: construct the factorized expression:
  // loop over number of factors, for each poly. factor
  // extract the coefficients
  // get power of l-th factor
  // get coefficients of l-th factor
  PolyOut=1;
  for (int l=0; l<factors.length(); ++l)
    { 
      IrreduxPoly=0;
      for (long i=0; i<=deg(factors[l].a); ++i)
	{
	  NumericFromZZ(PolyCoeff,factors[l].a.rep[i]);
	  IrreduxPoly+=PolyCoeff*pow(x,i);
	}
      PolyOut=PolyOut*(pow(IrreduxPoly,factors[l].b));
    }
  numeric coeff0;
  NumericFromZZ(coeff0, c);
  PolyOut=coeff0*PolyOut;

  return PolyOut ;
}


//////////////////////////////////////////////////////////////////////
// ZZFromNumeric: conversion from a GiNac integer to a NTL integer

void ZZFromNumeric(NTL::ZZ& zx, numeric& nx)
{
  // check if nx is an integer
  if (!nx.is_integer()) 
    { 
      throw(std::invalid_argument("second argument must be an integer"));
    }

  // if nx=0 set zx=0
  if(nx.is_zero())
    { 
      zx = 0; 
      return;
    }

  // if LONG_MIN<nx<LONG_MAX do direct conversion
  if ( nx<LONG_MAX && nx>LONG_MIN )
    {
      zx=nx.to_long(); 
      return;
    }

  // record sign
  bool sign= nx.is_positive();

  // make nx positive
  if (!sign) nx = -nx;

  // generate digits 
  std::string sx;
  numeric base(256);
  numeric remainder;
  long i = 0;
  do 
    {
      nx = iquo(nx, base, remainder); // cx[i]=nx%256, nx/=256;
      sx= sx + (char)remainder.to_int();
      i++;
    } 
  while (nx > 0);

  NTL::ZZFromBytes(zx, (const unsigned char *)sx.c_str() , i);
  if (!sign) zx = -zx;

  return;
}

//////////////////////////////////////////////////////////////////////
// NumericFromZZ: conversion from ntl integer to a GiNac integer 

void NumericFromZZ(numeric& nx, NTL::ZZ& zx)
{
// if zx=0 set nx=0
  if(zx==0)
    { 
      nx = 0; 
      return;
    }

  // if LONG_MIN<nx<LONG_MAX do direct conversion
  if ( zx<LONG_MAX && zx>LONG_MIN )
    {
      nx=to_long(zx); 
      return;
    }

  // record sign
  // make zx positive
  int sign=1;
  if(zx<0)
    {
      sign=-1; 
      zx=-zx;
    }

  // generate digits 
  int n=NumBytes(zx);
  unsigned char cx[n];
  NTL::BytesFromZZ(cx, zx, n);

  // calculate nx
  numeric base(256);
  numeric exponent;
  ex enx=0;
  for(int i=0; i<n; ++i)
    {
      exponent =i;
      enx += ((int)cx[i])*(pow(base,exponent));
    }

  enx = enx*sign;
  nx = ex_to<numeric>(enx);
  return;
}





