//
// integrate.cc
//

/*
 *  gTybalt Copyright (C) 2003 Stefan Weinzierl and Roberta Marani
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "integrate.h"

// global variables
#include "globals.h"

#include <cstdlib>
#include <iostream>
#include <fstream>
#include <string>
#include <sstream>
#include <vector>

#include <dlfcn.h>

#include <gsl/gsl_monte.h>
#include <gsl/gsl_monte_vegas.h>

extern "C" double fct_to_integrate(double *u, size_t dim, void *par);

// -----------------------------------------------------
// function of one variable

double intnum(GiNaC::ex expr, GiNaC::ex x, double xmin, double xmax)
{
  // prepare the expression
  GiNaC::symbol u0("u[0]");
  GiNaC::ex expr2 = expr.subs(GiNaC::lst{x==u0});

  double xl[1] = {xmin};
  double xu[1] = {xmax};

  return intnum_imp(expr2, 1, xl, xu);
}

// --------------------------------------------------------
// function of two variables

double intnum(GiNaC::ex expr, GiNaC::ex x, GiNaC::ex y, 
         double xmin, double xmax, double ymin, double ymax)
{
  // prepare the expression
  GiNaC::symbol u0("u[0]");
  GiNaC::symbol u1("u[1]");
  GiNaC::ex expr2 = expr.subs(GiNaC::lst{x==u0,y==u1});

  double xl[2] = {xmin,ymin};
  double xu[2] = {xmax,ymax};

  return intnum_imp(expr2, 2, xl, xu);
}

// --------------------------------------------------------
// function of n variables

double intnum(GiNaC::ex expr, GiNaC::lst x, GiNaC::lst xmin, GiNaC::lst xmax)
{
  int n = x.nops();

  std::vector<GiNaC::symbol> u;
  u.reserve(n);
  std::string name_start("u[");
  std::string name_end("]");

  for (int k=0; k<n; k++)
    {
      u.push_back( GiNaC::symbol( name_start+gtybalt_itos(k)+name_end ) );
    }

  GiNaC::ex expr2 = expr;
  for (int k=0; k<n; k++)
    {
      expr2 = expr2.subs( x.op(k) == u[k] );
    }

  double *xl = new double[n];
  double *xu = new double[n];
  for (int k=0; k<n; k++)
    {
      GiNaC::numeric xll = GiNaC::ex_to<GiNaC::numeric>(xmin.op(k));
      GiNaC::numeric xuu = GiNaC::ex_to<GiNaC::numeric>(xmax.op(k));
      xl[k] = xll.to_double();
      xu[k] = xuu.to_double();
    }

  double res = intnum_imp(expr2, n, xl, xu);

  // clean up
  delete[] xl;
  delete[] xu;

  return res;
}

// -----------------------------------------------------

double intnum_imp(GiNaC::ex expr, size_t dimension, double *xmin, double *xmax)
{
  typedef double entrypoint(double *uu, size_t ddim, void *ppar);
  entrypoint *module_fct = NULL;

  // write the function to a file 

  std::ofstream output_file;
  std::string file_name("xtybaltfct.c");

  output_file.open( file_name.c_str() );

  output_file << "// file automatically generated by gTybalt " << std::endl;
  output_file << std::endl;
  output_file << "#include <stddef.h> " << std::endl;
  output_file << "#include <stdlib.h> " << std::endl;
  output_file << "#include <math.h> " << std::endl;
  output_file << std::endl;
  output_file << "double fct_to_integrate(double *u, size_t dim, void *par)" << std::endl;
  output_file << "{" << std::endl;

  output_file << " double erg = " ;
  expr.print(GiNaC::print_csrc_double(output_file));
  output_file << ";" << std::endl;

  output_file << " return(erg); " << std::endl;
  output_file << "}" << std::endl;

  output_file.close();


  // compile
  system("gtybalt-compile xtybaltfct");

  // load the module
  void *module = NULL;
  module = dlopen("./xtybaltfct.so", RTLD_NOW);
  if (!module)
    {
      std::cout << "opening of module failed." << std::endl;
    }

  // find the entry point
  module_fct = (entrypoint*) dlsym(module,"fct_to_integrate");

  // integrate by Monte Carlo
  double res,err;

  const gsl_rng_type *T;
  gsl_rng *r;
  gsl_rng_env_setup();
  T = gsl_rng_default;
  r = gsl_rng_alloc(T);

  gsl_monte_function F = { module_fct, dimension, NULL};

  gsl_monte_vegas_state *mc_state = gsl_monte_vegas_alloc(dimension);

  // warm up to fix the grid
  mc_state->stage = 0;
  mc_state->iterations = gTybalt_int_iter_low; 
  gsl_monte_vegas_integrate(&F, xmin, xmax, dimension, 
			    gTybalt_int_calls_low, r, mc_state, 
			    &res, &err);

  // the main run
  mc_state->stage = 1;
  mc_state->iterations = gTybalt_int_iter_high; 
  gsl_monte_vegas_integrate(&F, xmin, xmax, dimension, 
			    gTybalt_int_calls_high, r, mc_state, 
			    &res, &err);

  // store the results in globals
  gTybalt_int_res  = res;
  gTybalt_int_err  = err;
  gTybalt_int_chi2 = mc_state->chisq;

  // clean up
  gsl_monte_vegas_free(mc_state);

  dlclose(module);

  system("rm xtybaltfct.c");
  system("rm xtybaltfct.o");
  system("rm xtybaltfct.so");

  return res;
}


/*
 * convert an integer into a string
 */
std::string gtybalt_itos(int arg)
{
    std::ostringstream buffer;
    buffer << arg; // send the int to the ostringstream
    return buffer.str(); // capture the string
}

